﻿' 
'    Copyright (c) 2011 Microsoft Corporation.  All rights reserved.
'    Use of this sample source code is subject to the terms of the Microsoft license 
'    agreement under which you licensed this sample source code and is provided AS-IS.
'    If you did not accept the terms of the license agreement, you are not authorized 
'    to use this sample source code.  For the terms of the license, please see the 
'    license agreement between you and Microsoft.
'  
'    To see the MSDN article about this app, visit http://go.microsoft.com/fwlink/?LinkId=247592 
'  
'

Imports System.Collections.ObjectModel
Imports System.ServiceModel.Syndication
Imports System.Text
Imports System.Text.RegularExpressions
Imports System.Xml.Linq

Public Class SynFeedParser
    Implements IXmlFeedParser

    ''' <summary>
    ''' Moderates thread operations.
    ''' </summary>
    Private Shared _lockObject As New Object

    ''' <summary>
    ''' Maintains the date of the latest article.
    ''' </summary>
    Public Shared latestDate As Date

    ''' <summary>
    ''' Default constructor
    ''' </summary>
    Public Sub New()
    End Sub

    ''' <summary>
    ''' Remove all HTTP tags and trim extra whitespace.
    ''' </summary>
    ''' <param name="text">The string from which HTML will be removed</param>
    ''' <returns>The HTML-removed string</returns>
    Private Shared Function HTMLParser(ByVal text As String) As String
        Dim parsedText As String = Regex.Replace(text, "</?\w+((\s+\w+(\s*=\s*(?:"".*?""|'.*?'|[^'"">\s]+))?)+\s*|\s*)/?>", String.Empty)
        parsedText = Regex.Replace(parsedText, "<!--.*-->", String.Empty)
        parsedText = HttpUtility.HtmlDecode(parsedText)
        parsedText = parsedText.Trim
        Return parsedText
    End Function

    ''' <summary>
    ''' Function to parse the item of a feed and obtain useful data.
    ''' <param name="feed">The Feed the article being parsed belongs to.</param>
    ''' <param name="synFeed">Syndication Feed to be parsed</param>
    ''' <param name="item">SyndicationItem that will be added to the database</param>
    ''' </summary>

    Public Function ParseItems(ByVal reader As XmlReader, ByVal parentFeed As Feed) As Collection(Of Article) Implements IXmlFeedParser.ParseItems
        ' Collection to store all parsed articles
        Dim parsedArticles As New Collection(Of Article)
        Dim synFeed As SyndicationFeed

        ' Create a syndicationFeed from the provided reader
        Try
            synFeed = SyndicationFeed.Load(reader)
        Catch
            Return Nothing
        End Try
        If Nothing IsNot synFeed Then
            ' First, obtain an image for the feed if it does not have one.
            If Not ImageGrabber.IfImageExists(parentFeed) Then
                ImageGrabber.GetImage(parentFeed, synFeed)
                latestDate = New Date
            End If

            ' Parse the xml file, getting each article.
            For Each item In synFeed.Items
                ' Get the necessary details from the SyndicationItem.
                Dim title As String = If(item.Title IsNot Nothing, item.Title.Text, String.Empty)
                title = HTMLParser(title) ' parse the title for html leftovers
                Dim itemURI As String = If(item.Links(0).Uri.ToString IsNot Nothing, item.Links(0).Uri.ToString, String.Empty)
                Dim [date] As DateTimeOffset = item.PublishDate

                ' Special case for the text/preview of article.
                Dim text As New StringBuilder
                If item.Summary IsNot Nothing Then
                    text.Append(item.Summary.Text)
                    ' If the article instead writes to the content:encoded portion of the xml.
                Else
                    For Each extension In item.ElementExtensions
                        Dim ele As XElement = extension.GetObject(Of XElement)()
                        If ele.Name.LocalName = "encoded" AndAlso ele.Name.Namespace.ToString.Contains("content") Then
                            text.Append(ele.Value & "<br/>")
                        End If
                    Next extension
                End If
                Dim parsedText As String = HTMLParser(text.ToString)
                If parsedText.Length >= 3500 Then
                    parsedText = parsedText.Substring(0, 2499)
                End If

                If Nothing = latestDate Then
                    SyncLock _lockObject
                        latestDate = New Date
                    End SyncLock
                End If

                Dim val As Boolean
                SyncLock _lockObject
                    val = latestDate.CompareTo([date].DateTime) < 0
                End SyncLock
                If val Then
                    ' Initialize new article, then add it to the database.
                    Dim newArticle As Article = New Article With {.ArticleTitle = title, .Authors = parentFeed.FeedTitle, .PublishDate = [date].DateTime.ToLocalTime(), .Summary = parsedText, .LastUpdatedTime = Date.Now, .ArticleBaseURI = itemURI, .FeedID = parentFeed.FeedID, .ImageURL = parentFeed.ImageURL}

                    parsedArticles.Add(newArticle)
                End If
            Next item
        End If
        Return parsedArticles
    End Function
End Class
