﻿' 
'    Copyright (c) 2011 Microsoft Corporation.  All rights reserved.
'    Use of this sample source code is subject to the terms of the Microsoft license 
'    agreement under which you licensed this sample source code and is provided AS-IS.
'    If you did not accept the terms of the license agreement, you are not authorized 
'    to use this sample source code.  For the terms of the license, please see the 
'    license agreement between you and Microsoft.
'  
'    To see the MSDN article about this app, visit http://go.microsoft.com/fwlink/?LinkId=247592 
'  
'
Imports System.Collections.Generic
Imports System.Collections.ObjectModel
Imports Microsoft.Phone.Net.NetworkInformation
Imports System.Threading

''' <summary>
''' Handles downloading of feeds and search results.
''' </summary>
Public NotInheritable Class WebTools
    ''' <summary>
    ''' Dictionary that stores all the downloaded feeds to a list of their respective articles.
    ''' </summary>
    Private _downloads As IDictionary(Of Feed, ICollection(Of Article))
    Public Property Downloads As IDictionary(Of Feed, ICollection(Of Article))
        Get
            Return _downloads
        End Get
        Private Set(ByVal value As IDictionary(Of Feed, ICollection(Of Article)))
            _downloads = value
        End Set
    End Property

    ''' <summary>
    ''' Dictionary storing all standing web requests.
    ''' </summary>
    Private _allRequests As IDictionary(Of Feed, WebRequest)

    ''' <summary>
    ''' Parser used to parse SyndicationItems into Articles.
    ''' </summary>
    Private _parser As IXmlFeedParser

    ''' <summary>
    ''' Timer used to count down until web requests are timed out.
    ''' </summary>
    Private _timeout As Timer

    ''' <summary>
    ''' Object used to lock an action.
    ''' </summary>
    Private _lockObject As Object

    ''' <summary>
    ''' Number of requests remaining to return.
    ''' </summary>
    Private Shared _numOfRequests As Integer

    ''' <summary>
    ''' Whether the application has been set to use only wifi connections.
    ''' </summary>
    Public Property WifiOnly As Boolean
        Get
            Return Settings.WifionlySetting
        End Get
        Set(ByVal value As Boolean)
            Settings.WifionlySetting = value
        End Set
    End Property

    ''' <summary>
    ''' Specifies if this WebTools Object is currently downloading.
    ''' This could have changed by the time you checked.
    ''' </summary>
    Public Property IsDownloading As Boolean

    ''' <summary>
    ''' Property for the DateTime of the last time Downloading was finished.
    ''' </summary>
    Public Shared ReadOnly LastDownloadProperty As DependencyProperty = DependencyProperty.Register("LastDownload", GetType(Date), GetType(WebTools), New PropertyMetadata(Nothing))

    ''' <summary>
    ''' Event raised after all feed downloads have finished.
    ''' </summary>
    Public Event AllDownloadsFinished As EventHandler(Of AllDownloadsFinishedEventArgs)

    ''' <summary>
    ''' Event raised after each feed download has finished.
    ''' </summary>
    Public Event SingleDownloadFinished As EventHandler(Of SingleDownloadFinishedEventArgs)

    Public Sub New(ByVal parser As IXmlFeedParser)
        If Nothing IsNot parser Then
            _parser = parser
        Else
            Throw New ArgumentNullException("No parser")
        End If

        ' Set all fields and properties.
        _lockObject = New Object
        _allRequests = New Dictionary(Of Feed, WebRequest)
        Downloads = New Dictionary(Of Feed, ICollection(Of Article))
    End Sub

    ''' <summary>
    ''' Downloads the provided Feed.
    ''' You cannot call download while downloads are in progress.
    ''' </summary>
    ''' <param name="feed">feed to download.</param>
    Public Sub Download(ByVal feed As Feed)
        If Nothing IsNot feed Then
            Download(New List(Of Feed) From {feed})
        End If
    End Sub

    ''' <summary>
    ''' Downloads the provided List of Feeds.
    ''' You cannot call download while downloads are in progress.
    ''' </summary>
    ''' <param name="feeds">List of feeds to download.</param>
    Public Sub Download(ByVal feeds As IList(Of Feed))
        ' Clear any leftover stored articles/feeds in Downloads and _allRequests.
        Downloads.Clear()
        _allRequests.Clear()

        ' Set the timer for timeouts if parsing syndication feeds.
        If TypeOf _parser Is SynFeedParser Then
            Dim timeLimit As Integer
            If Settings.InitialLaunchSetting Then
                timeLimit = 20000 ' 20 second limit for initial launch.
            Else
                timeLimit = 7500 ' Otherwise, 7.5 second time limit.
            End If
            Dim tc As New TimerCallback(AddressOf TimeoutConnections)
            _timeout = New Timer(tc, Me, timeLimit, Timeout.Infinite)
        End If

        If Nothing IsNot feeds AndAlso feeds.Count > 0 Then
            SyncLock _lockObject
                _numOfRequests += feeds.Count
            End SyncLock
            IsDownloading = True

            ' Download each separate feed.
            For Each feed In feeds
                Dim feedURI As String = feed.FeedBaseURI
                If Nothing IsNot feedURI Then
                    Dim feedRequest As HttpWebRequest = TryCast(HttpWebRequest.Create(feedURI), HttpWebRequest)

                    ' Check if the application has been set to wifi-only or not.
                    If WifiOnly Then
                        feedRequest.SetNetworkRequirement(NetworkSelectionCharacteristics.NonCellular)
                    End If
                    If Nothing IsNot feedRequest Then
                        Dim feedState As New RequestState With {.Request = feedRequest, .Argument = feed}
                        ' Change the owner to the parent HTTPWebRequest.
                        ' Change the argument to be the current feed to be downloaded.

                        ' Begin download.
                        feedRequest.BeginGetResponse(AddressOf ResponseCallback, feedState)

                        ' Keep track of the download.
                        _allRequests.Add(feed, feedRequest)
                    End If
                End If
            Next feed
        End If
    End Sub

    ''' <summary>
    ''' Callback method called when the "Download" method returns from an HTTPWebRequest.
    ''' </summary>
    ''' <param name="result">The result of the asynchronous operation.</param>
    Private Sub ResponseCallback(ByVal result As IAsyncResult)
        Dim state As RequestState = TryCast(result.AsyncState, RequestState)
        Dim parentFeed As Feed = TryCast(state.Argument, Feed)

        If Nothing IsNot parentFeed Then
            Dim request As HttpWebRequest = TryCast(state.Request, HttpWebRequest)

            ' Progress only if this download has not been timed out.

            If Nothing IsNot request Then
                ' Retrieve response.
                Try
                    Using response As HttpWebResponse = TryCast(request.EndGetResponse(result), HttpWebResponse)
                        If Nothing IsNot response AndAlso response.StatusCode = HttpStatusCode.OK Then
                            Using reader As XmlReader = XmlReader.Create(response.GetResponseStream)
                                ' Collection to store all articles
                                Dim parsedArticles As Collection(Of Article) = _parser.ParseItems(reader, parentFeed)

                                ' Raise event for a single feed downloaded.
                                If Nothing IsNot SingleDownloadFinishedEvent Then
                                    RaiseEvent SingleDownloadFinished(Me, New SingleDownloadFinishedEventArgs(parentFeed, parsedArticles))
                                End If

                                ' Add to all downloads dictionary and raise AllDownloadsFinished if all async requests have finished.
                                Downloads.Add(parentFeed, parsedArticles)
                                CheckIfDone()
                            End Using
                        End If
                    End Using
                Catch we As WebException
                    If we.Status = WebExceptionStatus.RequestCanceled Then
                        ' The web request was timed out. Let debug know it failed.
                        System.Diagnostics.Debug.WriteLine("ERROR: Feed """ & parentFeed.FeedTitle & """ was timed out.")
                        CheckIfDone()
                        Exit Sub
                    ElseIf we.Message = "The remote server returned an error: NotFound." Then
                        ' The web site did not respond. This means one of two things: Either the web site is bad, or you have no internet.
                        ' If error code NotFound is received, then there is no connection. Throw the exception.
                        If (TryCast(we.Response, HttpWebResponse)).StatusCode = HttpStatusCode.NotFound Then
                            ' WebTools throws an exception if without connection. Warn the user.
                            CheckIfDone()
                            If Not IsDownloading Then
                                Deployment.Current.Dispatcher.BeginInvoke(Sub() MessageBox.Show("FeedCast is unable to reach a connection. Please check your network connectivity."))
                            End If
                            ' If error code InternalServerError is received, then it is a faulty site.
                        Else
                            System.Diagnostics.Debug.WriteLine("ERROR: Feed """ & parentFeed.FeedTitle & """ is faulty.")
                            CheckIfDone()
                            Exit Sub
                        End If
                    Else
                        ' Unexpected web exception.
                        Throw we
                    End If
                End Try
            End If
        End If
    End Sub

    ''' <summary>
    ''' Decrements the number of remaining downloads and checks to see if all downloads are done.
    ''' <returns>If all downloads are finished</returns>
    ''' </summary>
    Private Sub CheckIfDone()
        SyncLock _lockObject
            _numOfRequests -= 1
            IsDownloading = ((_numOfRequests) > 0)
            If _numOfRequests <= 0 Then
                If Nothing IsNot AllDownloadsFinishedEvent Then
                    ' Dispose of the timer if using one.
                    Deployment.Current.Dispatcher.BeginInvoke(Sub()
                                                                  If TypeOf _parser Is SynFeedParser Then
                                                                      _timeout.Dispose()
                                                                  End If
                                                                  RaiseEvent AllDownloadsFinished(Me, New AllDownloadsFinishedEventArgs(Downloads))
                                                              End Sub)
                End If
            End If
        End SyncLock
    End Sub

    ''' <summary>
    ''' Aborts any web requests that are taking a suspiciously long amount of time.
    ''' <param name="allRequests">Object containing the current state</param>
    ''' </summary>
    Private Sub TimeoutConnections(ByVal state As Object)
        If IsDownloading Then
            System.Diagnostics.Debug.WriteLine("TIMEOUT BEGINNING.")

            For Each feed In _allRequests.Keys
                ' If the feed has not yet been downloaded, then cancel its web request.
                If Not Downloads.ContainsKey(feed) Then
                    _allRequests(feed).Abort()
                End If
            Next feed

            System.Diagnostics.Debug.WriteLine("TIMEOUT ENDING.")
        End If
    End Sub
End Class

''' <summary>
''' Class that passes data across asynchronous calls.
''' </summary>
Public Class RequestState
    ''' <summary>
    ''' The owner of this request state.
    ''' </summary>
    Public Property Request As WebRequest

    ''' <summary>
    ''' The argument for this request.
    ''' </summary>
    Public Property Argument As Object

    ''' <summary>
    ''' Creates a new, empty request state with default porperty values.
    ''' Request = null
    ''' Argument = null
    ''' AddToDatabase = null
    ''' AddToUI = null
    ''' </summary>
    Public Sub New()
        Me.Request = Nothing
        Me.Argument = Nothing
    End Sub

    ''' <summary>
    ''' Creates an instance of the webrequest class with given parameters.
    '''   <param name="argument">The argument for this request.</param>
    ''' <param name="addToDatabase">Boolean denoting whether to add the argument to the database.</param>
    ''' <param name="addToUI">Boolean denoting whether to add the argument to the UI.</param>
    ''' </summary>
    Public Sub New(ByVal argument As Object)
        Me.Argument = argument
    End Sub

    ''' <summary>
    ''' Creates an instance of the webrequest class with given parameters.
    '''  <param name="owner">The owner of this request state.</param>
    ''' <param name="argument">The argument for this request.</param>
    ''' <param name="addToDatabase">Boolean denoting whether to add the argument to the database.</param>
    ''' <param name="addToUI">Boolean denoting whether to add the argument to the UI.</param>
    ''' </summary>
    Public Sub New(ByVal owner As WebRequest, ByVal argument As Object)
        Me.New(argument)
        Me.Request = owner
    End Sub
End Class

''' <summary>
''' Event arguments passed when all Feed downloads have finished.
''' </summary>
Public Class AllDownloadsFinishedEventArgs
    Inherits EventArgs

    ''' <summary>
    ''' A dictionary of Downloaded feeds to a list of its respective downloaded articles.
    ''' </summary>
    Public Property Downloads As IDictionary(Of Feed, ICollection(Of Article))

    ''' <summary>
    ''' Creates a new instance of DownloadFinishedEventArgs which
    ''' provides arguments relevant to the download performed.
    ''' </summary>
    Public Sub New()
        Me.New(Nothing)
    End Sub

    ''' <summary>
    ''' Creates a new instance of DownloadFinishedEventArgs which
    ''' provides arguments relevant to the download performed.
    ''' </summary>
    ''' <param name="downloads">A dictionary of Downloaded feeds 
    ''' to a list of its respective downloaded articles.</param>
    Public Sub New(ByVal downloads As IDictionary(Of Feed, ICollection(Of Article)))
        MyBase.New()
        Me.Downloads = downloads
    End Sub
End Class

''' <summary>
''' Event argument passed when a single Feed download has finished.
''' Warning, this event is raised many times (once per feed downloaded).
''' </summary>
Public Class SingleDownloadFinishedEventArgs
    Inherits EventArgs

    ''' <summary>
    ''' The feed whose articles were downloaded.
    ''' </summary>
    Public Property ParentFeed As Feed

    ''' <summary>
    ''' Articles corresponding to ParentFeed that were downloaded.
    ''' </summary>
    Public Property DownloadedArticles As IList(Of Article)

    ''' <summary>
    ''' Creates a new instance of SingleDownloadFinishedEventArgs which
    ''' provides arguments relevant to the download performed.
    ''' </summary>
    Public Sub New()
        Me.New(Nothing, Nothing)
    End Sub

    ''' <summary>
    ''' Creates a new instance of SingleDownloadFinishedEventArgs which
    ''' provides arguments relevant to the download performed.
    ''' <param name="downloadedFeeds">The feed whose articles were downloaded.</param>
    ''' <param name="downloadedArticles">Articles corresponding to ParentFeed that were downloaded.</param>
    ''' </summary>
    Public Sub New(ByVal parentFeed As Feed, ByVal downloadedArticles As IList(Of Article))
        Me.ParentFeed = parentFeed
        Me.DownloadedArticles = downloadedArticles
    End Sub
End Class
